import pytest

from delfies import Orientation
from delfies.SAM_utils import SoftclippedRead
from delfies.seq_utils import cyclic_shifts, randomly_substitute, rev_comp
from delfies.telomere_utils import (
    TELOMERE_SEQS,
    has_softclipped_telo_array,
    record_telomere_additions,
    setup_telomere_tents,
)
from tests import ClassWithTempFasta

DEFAULT_ALIGNED_SEQ = "ATGCAAACCCAAAATTGGA"
DEFAULT_TELO_DICT = TELOMERE_SEQS["Nematoda"]
DEFAULT_NON_TELO_UNIT_FORWARD = "AAAAAA"
DEFAULT_MIN_TELO_ARRAY_SIZE = 3
DEFAULT_FORWARD_TELO = DEFAULT_TELO_DICT[Orientation.forward]
DEFAULT_FORWARD_TELO_ARRAY = DEFAULT_FORWARD_TELO * DEFAULT_MIN_TELO_ARRAY_SIZE
DEFAULT_REVERSE_TELO_ARRAY = (
    DEFAULT_TELO_DICT[Orientation.reverse] * DEFAULT_MIN_TELO_ARRAY_SIZE
)


@pytest.fixture
def softclipped_read():
    return SoftclippedRead(
        sequence=DEFAULT_ALIGNED_SEQ,
        reference_name="scaffold_1",
        sc_ref=4,
        sc_query=len(DEFAULT_ALIGNED_SEQ),
        sc_length=0,
        orientation=Orientation.forward,
    )


class TestFindSoftClippedTeloArraysInReads:
    def test_softclipped_read_with_nontelomeric_3prime_softclips_is_not_recognised(
        self, softclipped_read
    ):
        added_softclips = DEFAULT_NON_TELO_UNIT_FORWARD * DEFAULT_MIN_TELO_ARRAY_SIZE
        softclipped_read.sequence += added_softclips
        assert not has_softclipped_telo_array(
            softclipped_read,
            Orientation.forward,
            DEFAULT_TELO_DICT,
            DEFAULT_MIN_TELO_ARRAY_SIZE,
            0,
        )

    def test_softclipped_read_with_telomeric_3prime_softclips_is_recognised(
        self, softclipped_read
    ):
        softclipped_read.sequence += DEFAULT_FORWARD_TELO_ARRAY
        assert has_softclipped_telo_array(
            softclipped_read,
            Orientation.forward,
            DEFAULT_TELO_DICT,
            DEFAULT_MIN_TELO_ARRAY_SIZE,
            0,
        )

    def test_softclipped_read_with_telomeric_5prime_softclips_is_recognised(
        self, softclipped_read
    ):
        softclipped_read.sequence = DEFAULT_REVERSE_TELO_ARRAY
        softclipped_read.sc_query = len(DEFAULT_REVERSE_TELO_ARRAY) - 1
        assert has_softclipped_telo_array(
            softclipped_read,
            Orientation.reverse,
            DEFAULT_TELO_DICT,
            DEFAULT_MIN_TELO_ARRAY_SIZE,
            0,
        )

    def test_softclipped_read_with_shifted_telomeric_softclips_is_recognised(
        self, softclipped_read
    ):
        """
        Telomeres could be added starting with any of the bases in the telomeric
        repeat unit. This tests checks these are tolerated
        """
        read_seq = softclipped_read.sequence
        for orientation in Orientation:
            telo_unit = DEFAULT_TELO_DICT[orientation]
            for shifted_telo_unit in cyclic_shifts(telo_unit):
                added_softclips = (
                    shifted_telo_unit + shifted_telo_unit * DEFAULT_MIN_TELO_ARRAY_SIZE
                )
                if orientation is Orientation.forward:
                    softclipped_read.sequence = read_seq + added_softclips
                else:
                    softclipped_read.sequence = added_softclips + read_seq
                    softclipped_read.sc_query = len(added_softclips) - 1
                assert has_softclipped_telo_array(
                    softclipped_read,
                    orientation,
                    DEFAULT_TELO_DICT,
                    DEFAULT_MIN_TELO_ARRAY_SIZE,
                    0,
                )

    def test_softclipped_read_with_downstream_telomeric_softclips_is_not_recognised(
        self, softclipped_read
    ):
        """
        We only want to match telomere repeats found at the beginning of the
        softclipped portion of the `softclipped_read`.
        """
        read_seq = softclipped_read.sequence
        for orientation in Orientation:
            if orientation is Orientation.forward:
                non_telo_unit = DEFAULT_NON_TELO_UNIT_FORWARD
                added_softclips = (
                    non_telo_unit * DEFAULT_MIN_TELO_ARRAY_SIZE
                    + DEFAULT_FORWARD_TELO_ARRAY
                )
                softclipped_read.sequence = read_seq + added_softclips
            else:
                non_telo_unit = rev_comp(DEFAULT_NON_TELO_UNIT_FORWARD)
                added_softclips = (
                    DEFAULT_REVERSE_TELO_ARRAY
                    + non_telo_unit * DEFAULT_MIN_TELO_ARRAY_SIZE
                )
                softclipped_read.sequence = added_softclips + read_seq
                softclipped_read.sc_query = len(added_softclips) - 1
            assert non_telo_unit != DEFAULT_TELO_DICT[orientation]
            assert not has_softclipped_telo_array(
                softclipped_read,
                orientation,
                DEFAULT_TELO_DICT,
                DEFAULT_MIN_TELO_ARRAY_SIZE,
                0,
            )

    def test_softclipped_read_with_mutated_telomeres_is_recognised(
        self, softclipped_read
    ):
        mutated_telo_array = str()
        for _ in range(DEFAULT_MIN_TELO_ARRAY_SIZE):
            mutated_telo_array += randomly_substitute(
                DEFAULT_FORWARD_TELO, num_mutations=1
            )
        for orientation in Orientation:
            if orientation is Orientation.forward:
                softclipped_read.sequence += mutated_telo_array
            else:
                softclipped_read.sequence = (
                    rev_comp(mutated_telo_array) + softclipped_read.sequence
                )
                softclipped_read.sc_query = len(mutated_telo_array) - 1
            assert has_softclipped_telo_array(
                softclipped_read,
                orientation,
                DEFAULT_TELO_DICT,
                DEFAULT_MIN_TELO_ARRAY_SIZE,
                max_edit_distance=DEFAULT_MIN_TELO_ARRAY_SIZE,
            )

    def test_max_edit_distance_threshold_is_applied(self, softclipped_read):
        mutated_telo_array = str()
        for _ in range(DEFAULT_MIN_TELO_ARRAY_SIZE):
            mutated_telo_array += randomly_substitute(
                DEFAULT_FORWARD_TELO, num_mutations=1
            )
            softclipped_read.sequence += mutated_telo_array
            assert not has_softclipped_telo_array(
                softclipped_read,
                Orientation.forward,
                DEFAULT_TELO_DICT,
                DEFAULT_MIN_TELO_ARRAY_SIZE,
                max_edit_distance=1,
            )


class TestRecordTelomereAdditions(ClassWithTempFasta):
    def test_record_telomere_addition_forward(self, softclipped_read):
        reference = ">scaffold_1\nATGCTTTTT\n"
        softclipped_read.sc_query = 4
        fasta = self.make_fasta(reference)
        result = record_telomere_additions(
            [softclipped_read],
            telomere_unit_length=6,
            genome_fname=fasta,
            sample_name="test_sample",
        )
        expected = setup_telomere_tents()
        expected_tent = expected.new()
        expected_tent.update(
            contig="scaffold_1",
            start=softclipped_read.sc_ref - 1,
            end=softclipped_read.sc_ref,
            template_nucleotide="C",
            added_telomere=DEFAULT_ALIGNED_SEQ[
                softclipped_read.sc_query : softclipped_read.sc_query + 6
            ],
            sample_name="test_sample",
        )
        assert len(result) == 1
        assert result[0] == expected_tent

    def test_record_telomere_addition_reverse(self, softclipped_read):
        reference = ">scaffold_1\nCCCCCTTGGA\n"
        softclipped_read.sc_query = (
            len(DEFAULT_ALIGNED_SEQ) - 6
        )  # 1 + the length of the common sequence with 'DEFAULT_ALIGNED_SEQ', that, is 'TTGGA'
        softclipped_read.orientation = Orientation.reverse
        fasta = self.make_fasta(reference)
        result = record_telomere_additions(
            [softclipped_read],
            telomere_unit_length=6,
            genome_fname=fasta,
            sample_name="test_sample",
        )
        expected = setup_telomere_tents()
        expected_tent = expected.new()
        # Note the template_nucleotide and added_telomere are expressed in reverse complement
        # of the sequence in 'DEFAULT_ALIGNED_SEQ', due to read's reverse orientation.
        expected_tent.update(
            contig="scaffold_1",
            start=softclipped_read.sc_ref + 1,
            end=softclipped_read.sc_ref + 2,
            template_nucleotide="A",
            added_telomere="TTTTGG",
            sample_name="test_sample",
        )
        assert len(result) == 1
        assert result[0] == expected_tent



from pathlib import Path
from tempfile import TemporaryDirectory

from pyfastx import Fasta


class ClassWithTempFasta:
    @classmethod
    def setup_class(cls):
        cls.temp_dir = TemporaryDirectory()
        cls.temp_fasta = Path(cls.temp_dir.name) / "temp.fasta"

    @classmethod
    def teardown_class(cls):
        cls.temp_dir.cleanup()

    @classmethod
    def make_fasta(self, input_string, as_handle=False):
        with self.temp_fasta.open("w") as ofstream:
            ofstream.write(input_string)
        if as_handle:
            return Fasta(str(self.temp_fasta), build_index=True)
        else:
            return str(self.temp_fasta)



import pytest
from pybedtools import Interval as pybedtools_Interval

from delfies.interval_utils import Interval, get_contiguous_ranges, parse_region_string

EXAMPLE_INVALID_REGION_STRING = "chr1:2--200"
EXAMPLE_VALID_REGION_STRING = "chr1:2-200"


class TestRegionStrings:
    def test_invalid_region_strings_fail(self):
        with pytest.raises(ValueError):
            parse_region_string(EXAMPLE_INVALID_REGION_STRING)
        with pytest.raises(ValueError):
            parse_region_string("chr1::2-200")

    def test_valid_region_string(self):
        contig, start, stop = parse_region_string(EXAMPLE_VALID_REGION_STRING)
        assert contig == "chr1"
        assert start == 2
        assert stop == 200


class TestContiguousRanges:
    def test_get_repeated_range_for_single_input_num(self):
        input_nums = set([1])
        expected = [(1, 1)]
        assert get_contiguous_ranges(input_nums) == expected

    def test_contiguous_ranges_from_contiguous_inputs(self):
        input_nums = set([1, 2, 3, 4, 3, 2, 1])
        expected = [(1, 4)]
        assert get_contiguous_ranges(input_nums) == expected

    def test_contiguous_ranges_from_noncontiguous_inputs(self):
        input_nums = set([1, 2, 8, 7, 6, 5])
        expected = [(1, 2), (5, 8)]
        assert get_contiguous_ranges(input_nums) == expected


@pytest.fixture
def test_interval():
    return Interval("chr1", 2, 200)


class TestIntervalConstruction:
    def test_build_interval_from_invalid_region_string_fails(self):
        with pytest.raises(ValueError):
            Interval.from_region_string(EXAMPLE_INVALID_REGION_STRING)

    def test_build_interval_from_valid_region_string_succeeds(self, test_interval):
        result = Interval.from_region_string(EXAMPLE_VALID_REGION_STRING)
        assert result == test_interval

    def test_build_interval_from_pybedtools_interval(self, test_interval):
        pb_interval = pybedtools_Interval("chr1", 2, 200)
        result = Interval.from_pybedtools_interval(pb_interval)
        assert result == test_interval

    def test_interval_to_region_string(self, test_interval):
        assert test_interval.to_region_string() == EXAMPLE_VALID_REGION_STRING
        test_interval.start = None
        assert test_interval.to_region_string() == "chr1"


class TestIntervalCoordinates:
    def test_interval_with_and_without_coordinates(self, test_interval):
        assert test_interval.has_coordinates()
        test_interval.start = None
        assert not test_interval.has_coordinates()
        test_interval.start = 2
        test_interval.end = None
        assert not test_interval.has_coordinates()
        test_interval.start = None
        assert not test_interval.has_coordinates()

    def test_interval_spanning_invalid_coordinates(self, test_interval):
        with pytest.raises(ValueError):
            test_interval.start = None
            test_interval.spans(2)

    def test_interval_spanning_valid_coordinates(self, test_interval):
        assert test_interval.spans(test_interval.start)
        assert test_interval.spans(test_interval.start + 1)
        assert test_interval.spans(test_interval.end - 1)
        assert not test_interval.spans(test_interval.start - 1)
        assert not test_interval.spans(test_interval.end)


@pytest.fixture
def test_interval_two():
    return Interval("chr1", 50, 100)


class TestIntervalOverlaps:
    """
    Tests that Intervals are defined as [0-based inclusive, 0-based exclusive),
    and the difference between overlap (defined as at least one position in common between two intervals)
    and touching (defined as two contiguous overlaps).
    """

    def test_no_overlap_returns_false(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 400, 600)
        assert not test_interval_two.overlaps_or_touches(target_interval)

    def test_containment_overlap_returns_true(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 55, 80)
        assert test_interval_two.overlaps_or_touches(target_interval)

    def test_spanning_overlap_returns_true(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 1, 200)
        assert test_interval_two.overlaps_or_touches(target_interval)

    def test_overlap_but_wrong_name_returns_false(self, test_interval_two):
        target_interval = Interval(test_interval_two.name + "blah", 55, 80)
        assert not test_interval_two.overlaps_or_touches(target_interval)

    def test_zero_based_endpoint_overlap_returns_true(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 99, 150)
        assert test_interval_two.overlaps(target_interval)

    def test_zero_based_endpoint_touching_returns_true(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 100, 150)
        assert not test_interval_two.overlaps(target_interval)
        assert test_interval_two.overlaps_or_touches(target_interval)

    def test_one_based_endpoint_touching_returns_false(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 101, 150)
        assert not test_interval_two.overlaps_or_touches(target_interval)

    def test_zero_based_startpoint_overlap_returns_true(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 1, 51)
        assert test_interval_two.overlaps(target_interval)

    def test_zero_based_startpoint_touching_returns_true(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 1, 50)
        assert not test_interval_two.overlaps(target_interval)
        assert test_interval_two.overlaps_or_touches(target_interval)

    def test_one_based_startpoint_touching_returns_false(self, test_interval_two):
        target_interval = Interval(test_interval_two.name, 1, 49)
        assert not test_interval_two.overlaps_or_touches(target_interval)



from dataclasses import dataclass
from enum import Enum
from importlib import metadata
from typing import List, Tuple

from datasci import Tent

__version__ = metadata.version("delfies")

ID_DELIM = "__"
REGION_DELIM1 = ":"
REGION_DELIM2 = "-"
REGION_CLICK_HELP = (
    f"Region to focus on (format: 'contig{REGION_DELIM1}start{REGION_DELIM2}stop')"
)


class BreakpointType(Enum):
    """
    S2G: 'soma-to-germline': telomere-containing softclips aligned to non-telomere-containing genome region
    G2S: 'germline-to-soma': non-telomere-containing softclips aligned to telomere-containing genome region
    """

    S2G = "S2G"
    G2S = "G2S"

    def __str__(self):
        return f"{self.name}"


all_breakpoint_types = list(BreakpointType)


@dataclass
class BreakpointDetectionParams:
    bam_fname: str
    sample_name: str
    telomere_seqs: dict
    telo_array_size: int
    max_edit_distance: int
    clustering_threshold: int
    min_mapq: int
    read_filter_flag: int
    min_supporting_reads: int
    keep_telomeric_breakpoints: bool
    breakpoint_type: str = ""
    ofname_base: str = None


class Orientation(Enum):
    forward = "+"
    reverse = "-"


READ_SUPPORT_PREFIX = "num_supporting_reads"


@dataclass
class PutativeBreakpoint:
    orientation: Orientation
    max_value: int
    next_max_value: int
    max_value_other_orientation: int
    interval: Tuple[int, int]
    focus: Tent
    breakpoint_type: str = ""

    def update(self, query_focus: Tent):
        query_focus_value = int(
            query_focus[f"{READ_SUPPORT_PREFIX}{ID_DELIM}{self.orientation.name}"]
        )
        if query_focus_value > self.max_value:
            self.next_max_value = self.max_value
            self.max_value = query_focus_value
            self.focus = query_focus
        elif query_focus_value > self.next_max_value:
            self.next_max_value = query_focus_value


PutativeBreakpoints = List[PutativeBreakpoint]



from dataclasses import dataclass
from functools import reduce
from typing import List, Optional

from pysam import CSOFT_CLIP, AlignedSegment

from delfies import Orientation
from delfies.interval_utils import Interval, Intervals


@dataclass
class SoftclippedRead:
    """
    `sc`: 0-based position of softclip start, located at read extremity
    """

    sequence: str
    reference_name: str
    sc_ref: int
    sc_query: int
    sc_length: int
    orientation: Orientation


SoftclippedReads = List[SoftclippedRead]


_ordered_flags = [
    "PAIRED",
    "PROPER_PAIR",
    "UNMAP",
    "MUNMAP",
    "REVERSE",
    "MREVERSE",
    "READ1",
    "READ2",
    "SECONDARY",
    "QCFAIL",
    "DUP",
    "SUPPLEMENTARY",
]
FLAGS = {key: 2**x for x, key in enumerate(_ordered_flags)}
DEFAULT_READ_FILTER_NAMES = ["UNMAP", "SECONDARY", "DUP", "SUPPLEMENTARY"]
DEFAULT_READ_FILTER_FLAG = reduce(
    lambda x1, x2: x1 | x2, map(lambda el: FLAGS[el], DEFAULT_READ_FILTER_NAMES)
)
DEFAULT_MIN_MAPQ = 20


def read_flag_matches(read: AlignedSegment, filtering_SAM_flag: int) -> bool:
    return (read.flag & filtering_SAM_flag) != 0


def find_softclip_at_extremity(
    read: AlignedSegment, orientation: Orientation
) -> Optional[SoftclippedRead]:
    """
    In pysam (version 0.20.0), attributes `(reference|query_alignment)_start` and
    `(reference|query_alignment)_end` refer to positions of consumed alignment positions,
    i.e. excluding alignment softclips. `_start` refer to 0-based inclusive
    alignment-consuming positions on reference and query, and `_end` to 0-based
    exclusive alignment positions (i.e., 1 past the last alignment-consuming position).

    In the returned object, we return the position of the first softclipped position in
    both reference and read (query). If in forward orientation in the read,
    no adjustment is needed, and if in reverse orientation, we subtract one.
    """
    result = SoftclippedRead(
        read.query_sequence, read.reference_name, None, None, None, orientation
    )
    if orientation is Orientation.forward:
        if read.cigartuples[-1][0] == CSOFT_CLIP:
            result.sc_ref = read.reference_end
            result.sc_query = read.query_alignment_end
            result.sc_length = len(result.sequence) - result.sc_query
    else:
        if read.cigartuples[0][0] == CSOFT_CLIP:
            result.sc_ref = read.reference_start - 1
            result.sc_query = read.query_alignment_start - 1
            result.sc_length = result.sc_query + 1
    if result.sc_ref is None:
        return None
    else:
        return result


def filter_out_read_intervals(
    input_reads: SoftclippedReads, target_intervals: Intervals
) -> SoftclippedReads:
    result = list()
    for read in input_reads:
        if read.orientation is Orientation.forward:
            input_interval = Interval(read.reference_name, read.sc_ref - 1, read.sc_ref)
        else:
            input_interval = Interval(
                read.reference_name, read.sc_ref + 1, read.sc_ref + 2
            )
        keep_read = True
        for target_interval in target_intervals:
            if target_interval.overlaps(input_interval):
                keep_read = False
                break
        if keep_read:
            result.append(read)
    return result



import pytest

from delfies.interval_utils import Interval
from delfies.seq_utils import (
    cyclic_shifts,
    find_all_occurrences_in_genome,
    randomly_substitute,
    rev_comp,
)
from tests import ClassWithTempFasta


def test_rev_comp():
    seq1 = "AATTAACCGG"
    expected_seq1 = "CCGGTTAATT"
    assert rev_comp(seq1) == expected_seq1


class TestSequenceMutations:
    def test_not_a_nucleotide_fails(self):
        seq_to_mutate = "PPDD"
        with pytest.raises(ValueError):
            randomly_substitute(seq_to_mutate)

    def test_too_many_mutations_fails(self):
        seq_to_mutate = "AATT"
        with pytest.raises(ValueError):
            randomly_substitute(seq_to_mutate, num_mutations=len(seq_to_mutate) + 1)

    def test_various_numbers_of_mutations(self):
        seq_to_mutate = "AATTCCGG"
        num_nucleotides = len(seq_to_mutate)
        for num_mutations in range(num_nucleotides + 1):
            mutated_seq = randomly_substitute(seq_to_mutate, num_mutations)
            num_found_mutations = 0
            for i in range(num_nucleotides):
                if seq_to_mutate[i] != mutated_seq[i]:
                    num_found_mutations += 1
            assert num_found_mutations == num_mutations


def test_cyclic_shifts():
    str_to_shift = "TTAGGC"
    expected_shifts = [
        "TTAGGC",
        "TAGGCT",
        "AGGCTT",
        "GGCTTA",
        "GCTTAG",
        "CTTAGG",
    ]
    result = cyclic_shifts(str_to_shift)
    assert result == expected_shifts


class TestFindOccurrencesInGenomePerfectTeloArrays(ClassWithTempFasta):
    chrom_name = "chr1"
    telo_unit = "TTAGGC"
    default_query_array = telo_unit * 10
    telo_array = f"{rev_comp(default_query_array)}TT{default_query_array}"
    telo_array_record = f">{chrom_name}\n{telo_array}\n"
    search_region = [Interval(chrom_name)]
    targeted_region = [Interval(chrom_name, 0, len(default_query_array))]

    def test_find_all_occs_no_hits(self):
        fasta = self.make_fasta(self.telo_array_record)
        result = find_all_occurrences_in_genome(
            "AATTTTTTAAA", fasta, self.search_region, interval_window_size=0
        )
        assert result == []

    def test_find_all_occs_hits_whole_chrom(self):
        """
        Tests both lowercase and uppercase genomic nucleotides
        """
        all_fastas = [
            self.make_fasta(self.telo_array_record),
            self.make_fasta(f">{self.chrom_name}\n{self.telo_array.lower()}"),
        ]

        expected = [
            # Forward hit
            Interval(
                self.chrom_name,
                len(self.default_query_array) + 2,
                len(self.telo_array),
            ),
            # Reverse hit
            Interval(self.chrom_name, 0, len(self.default_query_array)),
        ]

        for fasta in all_fastas:
            result = find_all_occurrences_in_genome(
                self.default_query_array,
                fasta,
                self.search_region,
                interval_window_size=0,
            )
            assert result == expected

    def test_find_all_occs_hits_in_region(self):
        fasta = self.make_fasta(self.telo_array_record)
        result = find_all_occurrences_in_genome(
            self.default_query_array,
            fasta,
            self.targeted_region,
            interval_window_size=0,
        )
        expected = [
            Interval(
                self.chrom_name,
                self.targeted_region[0].start,
                self.targeted_region[0].end,
            )
        ]
        assert result == expected

    def test_find_all_occs_hits_with_window(self):
        fasta = self.make_fasta(self.telo_array_record)
        overflowing_window_size = 400
        result = find_all_occurrences_in_genome(
            self.default_query_array,
            fasta,
            self.targeted_region,
            interval_window_size=overflowing_window_size,
        )
        expected = [Interval(self.chrom_name, 0, len(self.telo_array))]
        assert result == expected


class TestFindOccurrencesInGenomeImperfectTeloArrays(ClassWithTempFasta):
    chrom_name = "chr1"
    telo_unit = "TTAGGC"
    mutated_telo_unit = "TTTGGC"
    telo_array = f"{telo_unit*3}{mutated_telo_unit*2}{telo_unit*3}"
    telo_array_record = f">{chrom_name}\n{telo_array}\n"
    search_region = [Interval(chrom_name)]

    def test_contiguous_units_are_clustered(self):
        fasta = self.make_fasta(self.telo_array_record)

        result = find_all_occurrences_in_genome(
            self.telo_unit,
            fasta,
            self.search_region,
            interval_window_size=0,
        )
        expected = [
            Interval(self.chrom_name, 0, len(self.telo_unit) * 3),
            Interval(self.chrom_name, len(self.telo_unit) * 5, len(self.telo_array)),
        ]
        assert result == expected

    def test_find_separate_arrays(self):
        fasta = self.make_fasta(self.telo_array_record)

        result = find_all_occurrences_in_genome(
            self.telo_unit * 3,
            fasta,
            self.search_region,
            interval_window_size=0,
        )

        expected = [
            Interval(self.chrom_name, 0, len(self.telo_unit) * 3),
            Interval(self.chrom_name, len(self.telo_unit) * 5, len(self.telo_unit) * 8),
        ]
        assert result == expected

    def test_cluster_arrays_within_window_size(self):
        fasta = self.make_fasta(self.telo_array_record)

        result = find_all_occurrences_in_genome(
            self.telo_unit * 3,
            fasta,
            self.search_region,
            interval_window_size=7,
        )

        expected = [
            Interval(self.chrom_name, 0, len(self.telo_unit) * 8),
        ]
        assert result == expected



from delfies import PutativeBreakpoint
from delfies.breakpoint_foci import setup_breakpoint_tents
from delfies.breakpoint_sequences import extract_breakpoint_sequences
from delfies.seq_utils import Orientation
from tests import ClassWithTempFasta


class TestExtractBreakpointSequences(ClassWithTempFasta):
    default_reference = ">scaffold_1\nACGTGATACA\n"
    tents = setup_breakpoint_tents()
    default_breakpoint = PutativeBreakpoint(Orientation.forward, 0, 0, 0, (0, 0), None)
    breakpoint_location = tents.new()
    breakpoint_location.update(contig="scaffold_1", start=3)
    default_breakpoint.focus = breakpoint_location

    def test_extract_breakpoint_within_boundaries_single_nucleotide(self):
        fasta = self.make_fasta(self.default_reference, as_handle=True)
        self.breakpoint_location.start = 3
        breakpoint_sequences = extract_breakpoint_sequences(
            [self.default_breakpoint], fasta, seq_window_size=1
        )
        assert len(breakpoint_sequences) == 1
        assert breakpoint_sequences[0].sequence == "GNT"

    def test_extract_breakpoint_within_boundaries_multiple_nucleotides(self):
        fasta = self.make_fasta(self.default_reference, as_handle=True)
        self.breakpoint_location.start = 3
        breakpoint_sequences = extract_breakpoint_sequences(
            [self.default_breakpoint], fasta, seq_window_size=2
        )
        assert len(breakpoint_sequences) == 1
        assert breakpoint_sequences[0].sequence == "CGNTG"

    def test_extract_breakpoint_outside_boundaries_negative_start(self):
        fasta = self.make_fasta(self.default_reference, as_handle=True)
        for start in [-2, 0]:
            self.breakpoint_location.start = start
            breakpoint_sequences = extract_breakpoint_sequences(
                [self.default_breakpoint], fasta, seq_window_size=2
            )
            assert len(breakpoint_sequences) == 1
            assert breakpoint_sequences[0].sequence == "NAC"

    def test_extract_breakpoint_outside_boundaries_end_past_last_nucleotide(self):
        fasta = self.make_fasta(self.default_reference, as_handle=True)
        self.breakpoint_location.start = 9
        breakpoint_sequences = extract_breakpoint_sequences(
            [self.default_breakpoint], fasta, seq_window_size=3
        )
        assert len(breakpoint_sequences) == 1
        assert breakpoint_sequences[0].sequence == "TACNA"

    def test_extract_breakpoint_within_boundaries_reverse_orientation(self):
        fasta = self.make_fasta(self.default_reference, as_handle=True)
        self.breakpoint_location.start = 3
        self.default_breakpoint.orientation = Orientation.reverse
        breakpoint_sequences = extract_breakpoint_sequences(
            [self.default_breakpoint], fasta, seq_window_size=2
        )
        assert len(breakpoint_sequences) == 1
        assert breakpoint_sequences[0].sequence == "CANCG"



from dataclasses import dataclass
from itertools import groupby
from operator import itemgetter
from typing import List, Optional, Set, Tuple

from delfies import REGION_DELIM1, REGION_DELIM2


@dataclass
class Interval:
    """
    Semi-closed intervals, as per BED standard:
        start: 0-based inclusive
        end: 0-based exclusive == 1-based inclusive
    """

    name: str
    start: Optional[int] = None
    end: Optional[int] = None

    def spans(self, query_pos: int) -> bool:
        if not self.has_coordinates():
            raise ValueError("Interval object has not been assigned coordinates")
        return self.start <= query_pos < self.end

    def overlaps(self, other: "Interval") -> bool:
        if not self.has_coordinates():
            raise ValueError("Interval object has not been assigned coordinates")
        return (
            (self.name == other.name)
            and (self.start < other.end)
            and (other.start < self.end)
        )

    def overlaps_or_touches(self, other: "Interval") -> bool:
        if not self.has_coordinates():
            raise ValueError("Interval object has not been assigned coordinates")
        return (
            (self.name == other.name)
            and (self.start <= other.end)
            and (other.start <= self.end)
        )

    def to_region_string(self) -> str:
        result = f"{self.name}"
        if self.has_coordinates():
            result += f"{REGION_DELIM1}{self.start}{REGION_DELIM2}{self.end}"
        return result

    @classmethod
    def from_region_string(cls, region_string: str) -> "Interval":
        contig, start, end = parse_region_string(region_string)
        return Interval(contig, start, end)

    @classmethod
    def from_pybedtools_interval(cls, pybedtools_interval) -> "Interval":
        return Interval(
            pybedtools_interval.chrom,
            pybedtools_interval.start,
            pybedtools_interval.end,
        )

    def has_coordinates(self) -> bool:
        return self.start is not None and self.end is not None


Intervals = List[Interval]


def parse_region_string(region_string: str) -> Tuple[str, int, int]:
    contig, regs = region_string.split(REGION_DELIM1)
    start, stop = map(lambda e: int(e.replace(",", "")), regs.split(REGION_DELIM2))
    return contig, start, stop


def get_contiguous_ranges(input_nums: Set[int]) -> List[Tuple[int, int]]:
    """
    Credit: https://stackoverflow.com/a/2154437/12519542
    """
    result = []

    for k, g in groupby(enumerate(sorted(input_nums)), lambda x: x[0] - x[1]):
        group = map(itemgetter(1), g)
        group = list(map(int, group))
        result.append((group[0], group[-1]))
    return result



import re
from dataclasses import dataclass
from random import choice as random_choice

from pyfastx import Fasta

from delfies import Orientation
from delfies.interval_utils import Interval, Intervals


@dataclass
class FastaRecord:
    ID: str
    sequence: str

    def __repr__(self):
        return f">{self.ID}\n{self.sequence}\n"


REVCOMP_TABLE_DNA = dict(A="T", C="G", G="C", T="A", N="N")
NUCLEOTIDES = set(REVCOMP_TABLE_DNA.keys())


def rev_comp(seq: str) -> str:
    result = "".join([REVCOMP_TABLE_DNA[elem] for elem in seq[::-1].upper()])
    return result


def randomly_substitute(seq: str, num_mutations: int = 1) -> str:
    if num_mutations > len(seq):
        raise ValueError(
            f"{num_mutations} is greater than the length of the input sequence"
        )
    index_choices = list(range(len(seq)))
    mutated_seq = seq
    while num_mutations != 0:
        index_choice = random_choice(index_choices)
        index_choices = [el for el in index_choices if el != index_choice]
        chosen_nucleotide = seq[index_choice]
        if chosen_nucleotide not in NUCLEOTIDES:
            raise ValueError("Not a nucleotide: {chosen_nucleotide}")
        possible_mutations = [el for el in NUCLEOTIDES if el != chosen_nucleotide]
        chosen_mutation = random_choice(possible_mutations)
        mutated_seq = (
            mutated_seq[:index_choice]
            + chosen_mutation
            + mutated_seq[index_choice + 1 :]
        )
        num_mutations -= 1
    return mutated_seq


def cyclic_shifts(input_str: str):
    result = list()
    for i in range(len(input_str)):
        result.append(input_str[i:] + input_str[:i])
    return result


def find_all_occurrences_in_genome(
    query_sequence: str,
    genome_fname: str,
    seq_regions: Intervals,
    interval_window_size: int,
) -> Intervals:
    genome_fasta = Fasta(genome_fname, build_index=True, uppercase=True)
    result = list()
    patterns = {
        Orientation.forward: re.compile(query_sequence),
        Orientation.reverse: re.compile(rev_comp(query_sequence)),
    }
    for seq_region in seq_regions:
        if seq_region.has_coordinates():
            relative_to_absolute = seq_region.start
            target_seq = genome_fasta[seq_region.name][
                seq_region.start : seq_region.end
            ]
        else:
            relative_to_absolute = 0
            target_seq = genome_fasta[seq_region.name]
        chrom_length = len(genome_fasta[seq_region.name])
        for pattern in patterns.values():
            match = None
            for match in pattern.finditer(str(target_seq)):
                new_interval = Interval(
                    name=seq_region.name,
                    start=max(
                        0, match.start() + relative_to_absolute - interval_window_size
                    ),
                    end=min(
                        chrom_length,
                        match.end() + relative_to_absolute + interval_window_size,
                    ),
                )
                if len(result) == 0:
                    result.append(new_interval)
                else:
                    if new_interval.overlaps_or_touches(result[-1]):
                        result[-1].end = new_interval.end
                    else:
                        result.append(new_interval)
            if match is not None and (
                len(result) == 0 or not new_interval.overlaps_or_touches(result[-1])
            ):
                result.append(new_interval)
    return result



import pytest
from pysam import AlignedSegment, AlignmentHeader

from delfies import BreakpointDetectionParams, BreakpointType, PutativeBreakpoint
from delfies.breakpoint_foci import (
    READ_SUPPORTS,
    FociWindow,
    Orientation,
    cluster_breakpoint_foci,
    record_softclips,
    setup_breakpoint_tents,
)
from delfies.interval_utils import Interval
from delfies.SAM_utils import DEFAULT_MIN_MAPQ, DEFAULT_READ_FILTER_FLAG
from delfies.seq_utils import rev_comp
from delfies.telomere_utils import TELOMERE_SEQS


@pytest.fixture
def breakpoint_focus():
    tents = setup_breakpoint_tents()
    new_tent = tents.new()
    new_tent.update(contig="test_contig", start=2, end=200)
    new_tent[f"{READ_SUPPORTS[0]}"] = 15
    new_tent[f"{READ_SUPPORTS[1]}"] = 20
    return new_tent


@pytest.fixture
def multiple_breakpoint_foci(breakpoint_focus):
    tents = setup_breakpoint_tents()
    tents.add(breakpoint_focus)
    new_tent = tents.new()
    new_tent.update(contig="test_contig", start=0, end=202)
    new_tent[f"{READ_SUPPORTS[0]}"] = 15
    tents.add(new_tent)
    new_tent = tents.new()
    new_tent.update(contig="test_contig", start=2000, end=2005)
    new_tent[f"{READ_SUPPORTS[0]}"] = 15
    tents.add(new_tent)
    return tents


@pytest.fixture
def focus_window():
    tents = setup_breakpoint_tents()
    new_tent = tents.new()
    new_tent.update(start=205, end=210)
    new_tent[f"{READ_SUPPORTS[0]}"] = 2
    new_tent[f"{READ_SUPPORTS[1]}"] = 200
    return FociWindow(new_tent)


@pytest.fixture
def putative_breakpoint():
    max_focus = PutativeBreakpoint(
        orientation=Orientation.forward,
        max_value=10,
        next_max_value=2,
        max_value_other_orientation=1,
        interval=(205, 210),
        focus=None,
    )
    return max_focus


class TestPutativeBreakpoint:
    def test_putative_breakpoint_update_no_new_max(
        self, breakpoint_focus, putative_breakpoint
    ):
        putative_breakpoint.max_value = 100
        putative_breakpoint.update(breakpoint_focus)
        assert putative_breakpoint.focus is None
        assert putative_breakpoint.max_value == 100
        assert putative_breakpoint.next_max_value == 15

    def test_putative_breakpoint_update_new_max(
        self, breakpoint_focus, putative_breakpoint
    ):
        prev_max_value = putative_breakpoint.max_value
        putative_breakpoint.update(breakpoint_focus)
        assert putative_breakpoint.focus == breakpoint_focus
        assert putative_breakpoint.max_value == breakpoint_focus[f"{READ_SUPPORTS[0]}"]
        assert putative_breakpoint.next_max_value == prev_max_value


class TestFociWindow:
    def test_inclusion_inside_or_overlapping_or_spanning(
        self, breakpoint_focus, focus_window
    ):
        # Overlapping
        fw_start = focus_window.foci[0].start
        fw_end = focus_window.foci[0].end
        breakpoint_focus.start = fw_start + 1
        breakpoint_focus.end = fw_end + 1
        assert focus_window.includes(breakpoint_focus, tolerance=0)
        breakpoint_focus.start = fw_start - 1
        breakpoint_focus.end = fw_end - 1
        # Inside
        assert focus_window.includes(breakpoint_focus, tolerance=0)
        breakpoint_focus.start = fw_start + 1
        breakpoint_focus.end = fw_end - 1
        assert focus_window.includes(breakpoint_focus, tolerance=0)
        # Spanning
        breakpoint_focus.start = fw_start - 1
        breakpoint_focus.end = fw_end + 1
        assert focus_window.includes(breakpoint_focus, tolerance=0)

    def test_no_inclusion_outside(self, breakpoint_focus, focus_window):
        fw_start = focus_window.foci[0].start
        fw_end = focus_window.foci[0].end
        breakpoint_focus.start = fw_start - 10
        breakpoint_focus.end = fw_start - 5
        assert not focus_window.includes(breakpoint_focus, tolerance=0)
        breakpoint_focus.start = fw_end + 10
        breakpoint_focus.end = fw_end + 15
        assert not focus_window.includes(breakpoint_focus, tolerance=0)

    def test_inclusion_with_tolerance(self, breakpoint_focus, focus_window):
        fw_start = focus_window.foci[0].start
        breakpoint_focus.start = fw_start - 2
        breakpoint_focus.end = fw_start - 1
        assert not focus_window.includes(breakpoint_focus, tolerance=0)
        assert focus_window.includes(breakpoint_focus, tolerance=1)

    def test_add(self, breakpoint_focus, focus_window):
        fw_start = focus_window.foci[0].start
        fw_end = focus_window.foci[0].end
        breakpoint_focus.start = fw_start + 1
        breakpoint_focus.end = fw_end + 2
        focus_window.add(breakpoint_focus)
        assert len(focus_window.foci) == 2
        assert focus_window.foci[1] == breakpoint_focus
        assert focus_window.Min == fw_start
        assert focus_window.Max == fw_end + 2

    def test_find_peak_softclip_focus_reverse_max(self, breakpoint_focus, focus_window):
        focus_window.add(breakpoint_focus)
        max_focus = focus_window.find_peak_softclip_focus()
        assert max_focus == PutativeBreakpoint(
            orientation=Orientation.reverse,
            max_value=200,
            next_max_value=20,
            max_value_other_orientation=15,
            interval=(focus_window.Min, focus_window.Max),
            focus=focus_window.foci[0],
        )

    def test_find_peak_softclip_focus_forward_max(self, breakpoint_focus, focus_window):
        breakpoint_focus[f"{READ_SUPPORTS[0]}"] = 400
        focus_window.add(breakpoint_focus)
        max_focus = focus_window.find_peak_softclip_focus()
        assert max_focus == PutativeBreakpoint(
            orientation=Orientation.forward,
            max_value=400,
            next_max_value=2,
            max_value_other_orientation=200,
            interval=(focus_window.Min, focus_window.Max),
            focus=breakpoint_focus,
        )


class TestBreakpointClustering:
    def test_breakpoint_foci_with_no_read_support_are_filtered_out(
        self, multiple_breakpoint_foci
    ):
        for focus in multiple_breakpoint_foci:
            focus[READ_SUPPORTS[0]] = 0
            focus[READ_SUPPORTS[1]] = 0
        result = cluster_breakpoint_foci(multiple_breakpoint_foci, tolerance=10)
        assert len(result) == 0

    def test_cluster_breakpoint_foci(self, multiple_breakpoint_foci):
        result = cluster_breakpoint_foci(multiple_breakpoint_foci, tolerance=10)
        assert len(result) == 2
        assert len(result[0].foci) == 2
        assert result[0].Min == 0
        assert result[0].Max == 202
        assert len(result[1].foci) == 1
        assert result[1].Min == 2000
        assert result[1].Max == 2005


DEFAULT_TELO_SEQ = TELOMERE_SEQS["Nematoda"][Orientation.forward]
DEFAULT_TELO_ARRAY_SIZE = 3
DEFAULT_NON_TELO_SEQ = "TAACCC"
DEFAULT_CHROM = "chr1"


@pytest.fixture
def read_telo_seq_forward_3prime():
    non_telo_seq = DEFAULT_NON_TELO_SEQ * 3
    telo_seq = DEFAULT_TELO_SEQ * DEFAULT_TELO_ARRAY_SIZE
    sequence = non_telo_seq + telo_seq
    cigar = ((0, len(non_telo_seq)), (4, len(telo_seq)))
    return (sequence, cigar)


@pytest.fixture
def read_telo_seq_reverse_3prime():
    non_telo_seq = DEFAULT_NON_TELO_SEQ * 3
    telo_seq = rev_comp(DEFAULT_TELO_SEQ) * DEFAULT_TELO_ARRAY_SIZE
    sequence = non_telo_seq + telo_seq
    cigar = ((0, len(non_telo_seq)), (4, len(telo_seq)))
    return (sequence, cigar)


@pytest.fixture
def softclipped_aligned_read():
    header = AlignmentHeader.from_dict(
        {
            "HD": {"VN": "1.0"},
            "SQ": [{"LN": 1e03, "SN": DEFAULT_CHROM}],
        }
    )
    read = AlignedSegment(header=header)
    read.reference_id = 0
    read.query_name = "simulated_read"
    read.flag = 0
    read.mapping_quality = DEFAULT_MIN_MAPQ
    read.reference_start = 3
    return read


@pytest.fixture
def detection_params():
    telomere_seqs = {
        Orientation.forward: DEFAULT_TELO_SEQ,
        Orientation.reverse: rev_comp(DEFAULT_TELO_SEQ),
    }
    return BreakpointDetectionParams(
        bam_fname="NA",
        sample_name=".",
        telomere_seqs=telomere_seqs,
        telo_array_size=DEFAULT_TELO_ARRAY_SIZE,
        max_edit_distance=0,
        clustering_threshold=1,
        min_mapq=DEFAULT_MIN_MAPQ,
        read_filter_flag=DEFAULT_READ_FILTER_FLAG,
        min_supporting_reads=1,
        keep_telomeric_breakpoints=False,
        breakpoint_type=BreakpointType.G2S,
    )


class TestRecordSoftclips:
    def test_G2S_read_with_3prime_forward_telo_softclips_is_rejected(
        self, detection_params, read_telo_seq_forward_3prime, softclipped_aligned_read
    ):
        softclipped_aligned_read.query_sequence = read_telo_seq_forward_3prime[0]
        softclipped_aligned_read.cigar = read_telo_seq_forward_3prime[1]
        breakpoint_tents = setup_breakpoint_tents()
        seq_region = Interval(DEFAULT_CHROM, 1, 100)
        record_softclips(
            softclipped_aligned_read,
            breakpoint_tents,
            dict(),
            detection_params,
            seq_region,
        )
        assert len(breakpoint_tents) == 0

    def test_G2S_read_with_3prime_reverse_telo_softclips_is_rejected(
        self, detection_params, read_telo_seq_reverse_3prime, softclipped_aligned_read
    ):
        softclipped_aligned_read.query_sequence = read_telo_seq_reverse_3prime[0]
        softclipped_aligned_read.cigar = read_telo_seq_reverse_3prime[1]
        breakpoint_tents = setup_breakpoint_tents()
        seq_region = Interval(DEFAULT_CHROM, 1, 100)
        record_softclips(
            softclipped_aligned_read,
            breakpoint_tents,
            dict(),
            detection_params,
            seq_region,
        )
        assert len(breakpoint_tents) == 0



from dataclasses import dataclass
from pathlib import Path
from random import choice as rand_choice
from tempfile import NamedTemporaryFile

import pytest
from pysam import AlignedSegment, AlignmentFile
from pysam import index as pysam_index
from pysam import qualitystring_to_array

from delfies import BreakpointType, Orientation
from delfies.breakpoint_foci import BreakpointDetectionParams, find_breakpoint_foci
from delfies.interval_utils import Interval
from delfies.SAM_utils import DEFAULT_MIN_MAPQ, DEFAULT_READ_FILTER_FLAG
from delfies.seq_utils import randomly_substitute, rev_comp
from delfies.telomere_utils import TELOMERE_SEQS

DEFAULT_NON_TELO_SEQ = "TAACCC"
DEFAULT_TELO_SEQ = TELOMERE_SEQS["Nematoda"][Orientation.forward]
DEFAULT_CHROM = "chr1"
DEFAULT_CHROM_LENGTH = 10e03
DEFAULT_MIN_TELO_ARRAY_SIZE = 10
DEFAULT_NUM_TELO_CONTAINING_READS = 150
EXPECTED_BREAKPOINT_POSITION = 1000


@dataclass
class TeloContainingReadGenerator:
    read_template: str = DEFAULT_NON_TELO_SEQ * 100 + DEFAULT_TELO_SEQ * 100
    num_reads: int = 200
    num_telo_containing_reads: int = DEFAULT_NUM_TELO_CONTAINING_READS
    telo_start_pos_in_genome: int = EXPECTED_BREAKPOINT_POSITION
    telo_start_pos_in_read: int = len(DEFAULT_NON_TELO_SEQ) * 100
    telo_array_size: int = len(DEFAULT_TELO_SEQ) * DEFAULT_MIN_TELO_ARRAY_SIZE * 2
    num_mutations: int = 0
    BAM_file: str = None

    def _add_telo_containing_softclipped_sequence(self, read: AlignedSegment) -> None:
        telo_start_pos = self.telo_start_pos_in_read
        index_start = rand_choice(range(telo_start_pos))
        non_telo_seq = self.read_template[index_start:telo_start_pos]
        telo_seq = self.read_template[
            telo_start_pos : telo_start_pos + self.telo_array_size
        ]
        read.reference_start = self.telo_start_pos_in_genome - len(non_telo_seq)
        read.cigar = ((0, len(non_telo_seq)), (4, len(telo_seq)))
        read.query_sequence = randomly_substitute(
            non_telo_seq, self.num_mutations
        ) + randomly_substitute(telo_seq, self.num_mutations)
        read.template_length = len(read.query_sequence)
        read.query_qualities = qualitystring_to_array("<" * len(read.query_sequence))

    def _add_nontelo_containing_fully_aligned_sequence(
        self, read: AlignedSegment
    ) -> None:
        telo_start_pos = self.telo_start_pos_in_read
        index_start = rand_choice(range(telo_start_pos))
        non_telo_seq_1 = self.read_template[index_start:telo_start_pos]
        non_telo_seq_2 = self.read_template[
            telo_start_pos - self.telo_array_size : telo_start_pos
        ]
        read.reference_start = self.telo_start_pos_in_genome - len(non_telo_seq_1)
        read.query_sequence = randomly_substitute(
            non_telo_seq_1, self.num_mutations
        ) + randomly_substitute(non_telo_seq_2, self.num_mutations)
        read.cigar = ((0, len(read.query_sequence)),)
        read.template_length = len(read.query_sequence)
        read.query_qualities = qualitystring_to_array("<" * len(read.query_sequence))

    def write_BAM(
        self,
        bitwise_flag: int = 0,
        mapping_quality: int = DEFAULT_MIN_MAPQ,
    ) -> str:
        header = {
            "HD": {"VN": "1.0"},
            "SQ": [{"LN": DEFAULT_CHROM_LENGTH, "SN": DEFAULT_CHROM}],
        }
        BAM_file = NamedTemporaryFile()
        with AlignmentFile(BAM_file.name, "wb", header=header) as outf:
            all_reads = list()
            for i in range(self.num_reads):
                read = AlignedSegment()
                read.reference_id = 0
                read.query_name = f"simulated_read_{i}"
                if i < self.num_telo_containing_reads:
                    self._add_telo_containing_softclipped_sequence(read)
                else:
                    self._add_nontelo_containing_fully_aligned_sequence(read)
                read.flag = bitwise_flag
                read.mapping_quality = mapping_quality
                all_reads.append(read)
            all_reads = sorted(all_reads, key=lambda el: el.reference_start)
            for read in all_reads:
                outf.write(read)
        self.BAM_file = BAM_file
        pysam_index(self.BAM_file.name)
        return self.BAM_file.name

    def __del__(self):
        if self.BAM_file is not None:
            self.BAM_file.close()
            Path(f"{self.BAM_file.name}.bai").unlink()


@pytest.fixture
def read_generator():
    return TeloContainingReadGenerator()


@pytest.fixture
def genome_interval():
    return Interval(DEFAULT_CHROM, 0, DEFAULT_CHROM_LENGTH)


@pytest.fixture
def detection_params():
    telomere_seqs = {
        Orientation.forward: DEFAULT_TELO_SEQ,
        Orientation.reverse: rev_comp(DEFAULT_TELO_SEQ),
    }
    return BreakpointDetectionParams(
        bam_fname="NA",
        sample_name=".",
        telomere_seqs=telomere_seqs,
        telo_array_size=DEFAULT_MIN_TELO_ARRAY_SIZE,
        max_edit_distance=0,
        clustering_threshold=1,
        min_mapq=DEFAULT_MIN_MAPQ,
        read_filter_flag=DEFAULT_READ_FILTER_FLAG,
        min_supporting_reads=10,
        keep_telomeric_breakpoints=False,
        breakpoint_type=BreakpointType.S2G,
    )


def test_forward_breakpoint_S2G(read_generator, genome_interval, detection_params):
    detection_params.bam_fname = read_generator.write_BAM()
    foci, _ = find_breakpoint_foci(detection_params, genome_interval)
    filtered_foci = [
        elem for elem in foci if elem.start == EXPECTED_BREAKPOINT_POSITION
    ]
    assert len(filtered_foci) == 1
    breakpoint_focus = filtered_foci[0]
    assert breakpoint_focus.contig == DEFAULT_CHROM
    assert breakpoint_focus.start == EXPECTED_BREAKPOINT_POSITION
    assert breakpoint_focus.end == EXPECTED_BREAKPOINT_POSITION + 1
    assert breakpoint_focus.breakpoint_type == str(BreakpointType.S2G)
    assert (
        breakpoint_focus.num_supporting_reads__forward
        == DEFAULT_NUM_TELO_CONTAINING_READS
    )



"""
Functions to find breakpoint foci. Definition: I call a breakpoint focus (plural: foci) a genomic
location at which 1+ reads are found bearing softclips compatible with a PDE breakpoint.
"""

from collections import defaultdict
from itertools import chain as it_chain
from typing import Dict, List, Tuple

from datasci import Tent, Tents
from pysam import AlignedSegment, AlignmentFile

from delfies import (
    ID_DELIM,
    READ_SUPPORT_PREFIX,
    BreakpointDetectionParams,
    BreakpointType,
    Orientation,
    PutativeBreakpoint,
)
from delfies.interval_utils import Interval, get_contiguous_ranges
from delfies.SAM_utils import (
    SoftclippedReads,
    find_softclip_at_extremity,
    read_flag_matches,
)
from delfies.telomere_utils import has_softclipped_telo_array

READ_SUPPORTS = [
    f"{READ_SUPPORT_PREFIX}{ID_DELIM}{o}" for o in map(lambda e: e.name, Orientation)
]


def setup_breakpoint_tents() -> Tents:
    tents_header = [
        "contig",
        "start",
        "end",
        "read_depth",
        "breakpoint_type",
        "sample_name",
    ] + READ_SUPPORTS
    tents = Tents(header=tents_header, required_header=tents_header[:5], unset_value=0)
    return tents


####################
## Foci detection ##
####################
def focus_has_enough_support(focus_tent, min_support: int) -> bool:
    result = False
    for read_support in READ_SUPPORTS:
        result |= int(focus_tent[read_support]) >= min_support
    return result


def record_softclips(
    aligned_read: AlignedSegment,
    breakpoint_foci: Tents,
    breakpoint_foci_positions,
    detection_params: BreakpointDetectionParams,
    seq_region: Interval,
) -> SoftclippedReads:
    reads_for_telomere_additions = list()
    for read_support in READ_SUPPORTS:
        orientation = Orientation[read_support.split(ID_DELIM)[1]]
        softclipped_read = find_softclip_at_extremity(aligned_read, orientation)
        if softclipped_read is None:
            continue
        if detection_params.breakpoint_type is BreakpointType.G2S:
            softclips_start_inside_target_region = seq_region.spans(
                softclipped_read.sc_ref
            )
            reject_softclipped_telo_array = False
            if not detection_params.keep_telomeric_breakpoints:
                # In G2S mode, we reject softclipped telomeres occurring in any orientation
                for G2S_tested_orientation in Orientation:
                    reject_softclipped_telo_array |= has_softclipped_telo_array(
                        softclipped_read,
                        G2S_tested_orientation,
                        detection_params.telomere_seqs,
                        min_telo_array_size=3,
                        max_edit_distance=detection_params.max_edit_distance,
                    )
            keep_read = (
                softclips_start_inside_target_region
                and not reject_softclipped_telo_array
            )
        else:
            softclipped_telo_array_found = has_softclipped_telo_array(
                softclipped_read,
                orientation,
                detection_params.telomere_seqs,
                detection_params.telo_array_size,
                max_edit_distance=detection_params.max_edit_distance,
            )
            keep_read = softclipped_telo_array_found
        if not keep_read:
            continue
        pos_to_commit = softclipped_read.sc_ref
        ref_name = aligned_read.reference_name
        match_tent_key = f"{ref_name}{ID_DELIM}{pos_to_commit}"
        if match_tent_key in breakpoint_foci_positions:
            breakpoint_foci_positions[match_tent_key][read_support] += 1
        else:
            new_tent = breakpoint_foci.new()
            new_tent.update(
                contig=ref_name,
                start=pos_to_commit,
                end=pos_to_commit + 1,
                breakpoint_type=str(detection_params.breakpoint_type),
                sample_name=detection_params.sample_name,
            )
            new_tent[read_support] += 1
            breakpoint_foci_positions[match_tent_key] = new_tent
        if detection_params.breakpoint_type is BreakpointType.S2G:
            reads_for_telomere_additions.append(softclipped_read)
    return reads_for_telomere_additions


def find_breakpoint_foci(
    detection_params: BreakpointDetectionParams,
    seq_region: Interval,
) -> Tuple[Tents, SoftclippedReads]:
    reads_for_telomere_additions = list()
    breakpoint_foci = setup_breakpoint_tents()
    breakpoint_foci_positions = {}
    contig_name = seq_region.name
    if seq_region.has_coordinates():
        fetch_args = dict(
            contig=contig_name, start=seq_region.start, stop=seq_region.end
        )
    else:
        fetch_args = dict(contig=contig_name)
    bam_fstream = AlignmentFile(detection_params.bam_fname)
    for aligned_read in bam_fstream.fetch(**fetch_args):
        if aligned_read.mapping_quality < detection_params.min_mapq:
            continue
        if read_flag_matches(aligned_read, detection_params.read_filter_flag):
            continue
        reads_for_telomere_additions.extend(
            record_softclips(
                aligned_read,
                breakpoint_foci,
                breakpoint_foci_positions,
                detection_params,
                seq_region,
            )
        )
    # Filter for minimum support
    breakpoint_foci_positions = {
        key: val
        for key, val in breakpoint_foci_positions.items()
        if focus_has_enough_support(val, detection_params.min_supporting_reads)
    }
    # Expand to a few positions before and after putative breakpoints: allows users to
    # assess changes in coverage around breakpoints (using the corresponding output tsv)
    positions_to_commit = set()
    for focus_tent in breakpoint_foci_positions.values():
        committed_position = focus_tent["start"]
        positions_to_commit.update(
            range(committed_position - 2, committed_position + 3)
        )
    record_read_depth_at_breakpoint_foci(
        positions_to_commit,
        breakpoint_foci_positions,
        contig_name,
        breakpoint_foci,
        bam_fstream,
        detection_params,
    )
    return breakpoint_foci, reads_for_telomere_additions


def record_read_depth_at_breakpoint_foci(
    positions_to_commit,
    breakpoint_foci_positions,
    contig_name,
    breakpoint_foci,
    bam_fstream,
    detection_params,
):
    """
    Adds read depth at each position in positions_to_commit,
    using pysam 'pileup'.
    Special cases:
        - When a breakpoint occurs at first position of a contig, 'pileup' will not
          record any read depth as the breakpoint position is set to -1.
          So we manually commit that position to :breakpoint_foci: to ensure it is output.
        - [TODO] When a breakpoint occurs at last position of a contig
    """
    for start, end in get_contiguous_ranges(positions_to_commit):
        if start < 0:
            negative_tent_key = f"{contig_name}{ID_DELIM}-1"
            if negative_tent_key in breakpoint_foci_positions:
                breakpoint_foci.add(breakpoint_foci_positions[negative_tent_key])
        # +1 for `end` because `end` needs to be exclusive in pysam `pileup`
        pileup_args = dict(
            contig=contig_name,
            start=max(start, 0),
            end=max(end + 1, 0),
            flag_filter=detection_params.read_filter_flag,
            min_mapping_quality=detection_params.min_mapq,
            ignore_orphans=False,
            truncate=True,
        )
        for pileup_column in bam_fstream.pileup(**pileup_args):
            read_depth = pileup_column.nsegments
            ref_pos = pileup_column.reference_pos
            tent_key = f"{contig_name}{ID_DELIM}{ref_pos}"
            if tent_key in breakpoint_foci_positions:
                breakpoint_foci_positions[tent_key]["read_depth"] = read_depth
                breakpoint_foci.add(breakpoint_foci_positions[tent_key])
            else:
                new_tent = breakpoint_foci.new()
                new_tent.update(
                    contig=contig_name,
                    start=ref_pos,
                    end=ref_pos + 1,
                    read_depth=read_depth,
                    breakpoint_type=str(detection_params.breakpoint_type),
                    sample_name=detection_params.sample_name,
                )
                breakpoint_foci.add(new_tent)


#####################
## Foci clustering ##
#####################
class FociWindow:
    def __init__(self, focus):
        self.foci = [focus]
        self.Min = int(focus.start)
        self.Max = int(focus.end)

    def includes(self, focus: Tent, tolerance: int):
        focus_start_past_end = int(focus.start) > self.Max + tolerance
        focus_end_before_start = int(focus.end) < self.Min - tolerance
        return not focus_start_past_end and not focus_end_before_start

    def add(self, focus):
        self.foci.append(focus)
        start = int(focus.start)
        end = int(focus.end)
        if end > self.Max:
            self.Max = end
        if start < self.Min:
            self.Min = start

    def find_peak_softclip_focus(self) -> PutativeBreakpoint:
        forward_maximum = PutativeBreakpoint(
            Orientation.forward, 0, 0, 0, (self.Min, self.Max), None
        )
        reverse_maximum = PutativeBreakpoint(
            Orientation.reverse, 0, 0, 0, (self.Min, self.Max), None
        )
        for focus in self.foci:
            forward_maximum.update(focus)
            reverse_maximum.update(focus)
        if forward_maximum.max_value > reverse_maximum.max_value:
            max_maximum = forward_maximum
            max_maximum.max_value_other_orientation = reverse_maximum.max_value
        else:
            max_maximum = reverse_maximum
            max_maximum.max_value_other_orientation = forward_maximum.max_value
        return max_maximum

    def __repr__(self):
        return f"[{self.Min},{self.Max}]"


def cluster_breakpoint_foci(foci: Tents, tolerance: int) -> List[FociWindow]:
    """
    Developer note:
        foci without any softclipped-reads are ignored for the purpose of clustering,
        as they are only present in the output tsv to assess coverage changes near breakpoints.
    """
    result: Dict[str, List[FociWindow]] = defaultdict(list)
    for focus in foci:
        if not focus_has_enough_support(focus, 1):
            continue
        contig_windows = result[focus.contig]
        found_window = False
        for elem in contig_windows:
            if elem.includes(focus, tolerance=tolerance):
                elem.add(focus)
                found_window = True
                break
        if not found_window:
            contig_windows.append(FociWindow(focus))
    return list(it_chain(*result.values()))



from datasci import Tents
from edlib import align as edlib_align
from pyfastx import Fasta

from delfies import Orientation
from delfies.SAM_utils import SoftclippedRead, SoftclippedReads
from delfies.seq_utils import rev_comp

TELOMERE_SEQS = {
    "Nematoda": {Orientation.forward: "TTAGGC", Orientation.reverse: "GCCTAA"}
}


def has_softclipped_telo_array(
    read: SoftclippedRead,
    orientation: Orientation,
    telomere_seqs,
    min_telo_array_size: int,
    max_edit_distance: int,
) -> bool:
    """
    Note: we allow for the softclipped telo array to start with any cyclic shift
    of the telomeric repeat unit.
    """
    telo_unit = telomere_seqs[orientation]
    searched_telo_array = telo_unit * min_telo_array_size
    subseq_clip_end = len(searched_telo_array) + len(telo_unit)
    if orientation is Orientation.forward:
        end = read.sc_query + subseq_clip_end
        subseq = read.sequence[read.sc_query : end]
    else:
        start = max(read.sc_query + 1 - subseq_clip_end, 0)
        subseq = read.sequence[start : read.sc_query + 1]
    result = edlib_align(
        searched_telo_array, subseq, mode="HW", task="distance", k=max_edit_distance
    )
    found_telo_array = result["editDistance"] != -1
    return found_telo_array


def setup_telomere_tents() -> Tents:
    tents_header = [
        "contig",
        "start",
        "end",
        "template_nucleotide",
        "added_telomere",
        "sample_name",
    ]
    tents = Tents(header=tents_header)
    return tents


def record_telomere_additions(
    reads_for_telomere_additions: SoftclippedReads,
    telomere_unit_length: int,
    genome_fname: str,
    sample_name: str,
) -> Tents:
    genome = Fasta(genome_fname)
    telomere_tents = setup_telomere_tents()
    for read in reads_for_telomere_additions:
        sc_query = read.sc_query
        if read.orientation is Orientation.forward:
            added_telomere = read.sequence[sc_query : sc_query + telomere_unit_length]
            ref_position_one_based = read.sc_ref
            template_nucleotide = genome.fetch(
                read.reference_name, (ref_position_one_based, ref_position_one_based)
            )
        else:
            added_telomere = rev_comp(
                read.sequence[sc_query + 1 - telomere_unit_length : sc_query + 1]
            )
            ref_position_one_based = read.sc_ref + 2
            template_nucleotide = rev_comp(
                genome.fetch(
                    read.reference_name,
                    (ref_position_one_based, ref_position_one_based),
                )
            )
        new_tent = telomere_tents.new()
        new_tent.update(
            contig=read.reference_name,
            start=ref_position_one_based - 1,
            end=ref_position_one_based,
            template_nucleotide=template_nucleotide,
            added_telomere=added_telomere,
            sample_name=sample_name,
        )
        telomere_tents.add(new_tent)
    return telomere_tents



import pytest
from pysam import CMATCH, CSOFT_CLIP, AlignedSegment, AlignmentHeader

from delfies import Orientation
from delfies.interval_utils import Interval
from delfies.SAM_utils import (
    FLAGS,
    SoftclippedRead,
    filter_out_read_intervals,
    find_softclip_at_extremity,
    read_flag_matches,
)

DEFAULT_ALIGNED_SEQ = "ATGCAAAAAAAAATTTGGA"
DEFAULT_SOFTCLIPPED_SEQ = "TATAGGTAACATCGCGGCATTCTACGG"
DEFAULT_REF_NAME = "test_reference"


@pytest.fixture
def pysam_basic_read():
    header = AlignmentHeader.from_references(
        reference_names=[DEFAULT_REF_NAME], reference_lengths=[100000]
    )
    aligned_seq = DEFAULT_ALIGNED_SEQ
    len_aligned_seq = len(aligned_seq)
    read = AlignedSegment(header=header)
    read.query_sequence = aligned_seq
    read.query_name = "test_query"
    read.reference_name = DEFAULT_REF_NAME
    read.reference_start = 200
    read.cigartuples = [(CMATCH, len_aligned_seq)]
    return read


@pytest.fixture
def pysam_read_with_3prime_softclips(pysam_basic_read):
    seq_added = DEFAULT_SOFTCLIPPED_SEQ
    read = pysam_basic_read
    read.query_sequence += seq_added
    read.cigartuples += [(CSOFT_CLIP, len(seq_added))]
    return read


@pytest.fixture
def pysam_read_with_5prime_softclips(pysam_basic_read):
    seq_added = DEFAULT_SOFTCLIPPED_SEQ
    read = pysam_basic_read
    read.query_sequence = seq_added + read.query_sequence
    read.cigartuples = [(CSOFT_CLIP, len(seq_added))] + read.cigartuples
    return read


class TestReadBitWiseFiltering:
    read_flags = FLAGS["UNMAP"] | FLAGS["DUP"]

    def test_read_flag_matches(self, pysam_basic_read):
        pysam_basic_read.flag = self.read_flags
        assert read_flag_matches(pysam_basic_read, FLAGS["PAIRED"] | FLAGS["UNMAP"])

    def test_read_flag_no_matches(self, pysam_basic_read):
        pysam_basic_read.flag = self.read_flags
        assert not read_flag_matches(
            pysam_basic_read, FLAGS["PAIRED"] | FLAGS["SUPPLEMENTARY"]
        )

    def test_read_flag_zero_no_matches(self, pysam_basic_read):
        # Flag of zero == read is mapped in forward orientation
        # On a read with such a flag, no filtering is applicable
        pysam_basic_read.flag = 0
        assert not read_flag_matches(pysam_basic_read, FLAGS["PAIRED"] | FLAGS["UNMAP"])


class TestSoftclipDetection:
    def test_read_no_softclips(self, pysam_basic_read):
        result_forward = find_softclip_at_extremity(
            pysam_basic_read, Orientation.forward
        )
        result_reverse = find_softclip_at_extremity(
            pysam_basic_read, Orientation.reverse
        )
        assert result_forward is None
        assert result_reverse is None

    def test_read_with_3prime_softclips(self, pysam_read_with_3prime_softclips):
        result_forward = find_softclip_at_extremity(
            pysam_read_with_3prime_softclips, Orientation.forward
        )
        assert result_forward == SoftclippedRead(
            sequence=pysam_read_with_3prime_softclips.query_sequence,
            reference_name=pysam_read_with_3prime_softclips.reference_name,
            sc_ref=pysam_read_with_3prime_softclips.reference_start
            + len(DEFAULT_ALIGNED_SEQ),
            sc_query=len(DEFAULT_ALIGNED_SEQ),
            sc_length=len(DEFAULT_SOFTCLIPPED_SEQ),
            orientation=Orientation.forward,
        )
        assert (
            find_softclip_at_extremity(
                pysam_read_with_3prime_softclips, Orientation.reverse
            )
            is None
        )

    def test_read_with_5prime_softclips(self, pysam_read_with_5prime_softclips):
        result_reverse = find_softclip_at_extremity(
            pysam_read_with_5prime_softclips, Orientation.reverse
        )
        assert result_reverse == SoftclippedRead(
            sequence=pysam_read_with_5prime_softclips.query_sequence,
            reference_name=pysam_read_with_5prime_softclips.reference_name,
            sc_ref=pysam_read_with_5prime_softclips.reference_start - 1,
            sc_query=len(DEFAULT_SOFTCLIPPED_SEQ) - 1,
            sc_length=len(DEFAULT_SOFTCLIPPED_SEQ),
            orientation=Orientation.reverse,
        )
        assert (
            find_softclip_at_extremity(
                pysam_read_with_5prime_softclips, Orientation.forward
            )
            is None
        )


@pytest.fixture
def basic_softclipped_read():
    read = SoftclippedRead(
        DEFAULT_ALIGNED_SEQ,
        DEFAULT_REF_NAME,
        sc_ref=400,
        sc_query=200,
        sc_length=50,
        orientation=Orientation.forward,
    )
    return read


class TestReadIntervalFiltering:
    default_range = [Interval(DEFAULT_REF_NAME, 100, 150)]

    def test_read_outside_range_not_filtered(self, basic_softclipped_read):
        result = filter_out_read_intervals([basic_softclipped_read], self.default_range)
        expected = [basic_softclipped_read]
        assert result == expected

    def test_inside_range_filtered(self, basic_softclipped_read):
        basic_softclipped_read.sc_ref = 101
        result = filter_out_read_intervals([basic_softclipped_read], self.default_range)
        expected = []
        assert result == expected

    def test_at_range_exclusion_point_not_filtered(self, basic_softclipped_read):
        """
        We want the ref position just before the softclipped position to be considered
        """
        basic_softclipped_read.sc_ref = 100
        result = filter_out_read_intervals([basic_softclipped_read], self.default_range)
        expected = [basic_softclipped_read]
        assert result == expected

    def test_at_range_exclusion_point_reverse_not_filtered(
        self, basic_softclipped_read
    ):
        """
        We want the ref position just after the softclipped position to be considered
        """
        basic_softclipped_read.orientation = Orientation.reverse
        basic_softclipped_read.sc_ref = 149
        result = filter_out_read_intervals([basic_softclipped_read], self.default_range)
        expected = [basic_softclipped_read]
        assert result == expected



from pathlib import Path
from typing import List

from pyfastx import Fasta

from delfies import Orientation, PutativeBreakpoints
from delfies.breakpoint_foci import READ_SUPPORT_PREFIX
from delfies.seq_utils import FastaRecord, rev_comp


def extract_breakpoint_sequences(
    maximal_foci: PutativeBreakpoints, genome: Fasta, seq_window_size: int
) -> List[FastaRecord]:
    result = list()
    for max_focus in maximal_foci:
        focus = max_focus.focus
        breakpoint_pos = max(int(focus.start), 0)
        windowed_start = max(breakpoint_pos - seq_window_size + 1, 0)
        windowed_end = breakpoint_pos + seq_window_size
        breakpoint_sequence = (
            genome.fetch(focus.contig, (windowed_start, breakpoint_pos))
            + "N"
            + genome.fetch(focus.contig, (breakpoint_pos + 1, windowed_end))
        )
        strand_name = "3prime"
        if max_focus.orientation is Orientation.reverse:
            breakpoint_sequence = rev_comp(breakpoint_sequence)
            strand_name = "5prime"
        breakpoint_name = f"{max_focus.breakpoint_type}_{strand_name}_{focus.contig} breakpoint_pos:{breakpoint_pos} {READ_SUPPORT_PREFIX}:{max_focus.max_value} next_best_value_on_same_strand:{max_focus.next_max_value} best_value_on_other_strand:{max_focus.max_value_other_orientation}"
        result.append(FastaRecord(breakpoint_name, breakpoint_sequence))
    return result


def write_breakpoint_sequences(
    genome_fname: str,
    maximal_foci: PutativeBreakpoints,
    odirname: Path,
    seq_window_size: int,
) -> None:
    genome = Fasta(genome_fname)
    breakpoint_sequences = extract_breakpoint_sequences(
        maximal_foci, genome, seq_window_size
    )
    breakpoint_fasta = odirname / "breakpoint_sequences.fasta"
    with breakpoint_fasta.open("w") as ofstream:
        for breakpoint_sequence in breakpoint_sequences:
            ofstream.write(str(breakpoint_sequence))



import itertools as it
import multiprocessing as mp
from pathlib import Path
from typing import Tuple

import rich_click as click
from datasci import Tents
from pybedtools import BedTool
from pysam import AlignmentFile

from delfies import (
    ID_DELIM,
    REGION_CLICK_HELP,
    BreakpointType,
    Orientation,
    PutativeBreakpoints,
    __version__,
    all_breakpoint_types,
)
from delfies.breakpoint_foci import (
    BreakpointDetectionParams,
    cluster_breakpoint_foci,
    find_breakpoint_foci,
    setup_breakpoint_tents,
)
from delfies.breakpoint_sequences import write_breakpoint_sequences
from delfies.interval_utils import Interval, Intervals
from delfies.SAM_utils import (
    DEFAULT_MIN_MAPQ,
    DEFAULT_READ_FILTER_FLAG,
    DEFAULT_READ_FILTER_NAMES,
    SoftclippedReads,
    filter_out_read_intervals,
)
from delfies.seq_utils import find_all_occurrences_in_genome, rev_comp
from delfies.telomere_utils import TELOMERE_SEQS, record_telomere_additions

click.rich_click.OPTION_GROUPS = {
    "delfies": [
        {
            "name": "Generic",
            "options": ["--help", "--version", "--threads", "--sample_name"],
        },
        {
            "name": "Region selection",
            "options": ["--seq_region", "--bed"],
        },
        {
            "name": "Breakpoint detection",
            "options": [
                "--breakpoint_type",
                "--telo_forward_seq",
                "--telo_array_size",
                "--telo_max_edit_distance",
                "--min_mapq",
                "--read_filter_flag",
                "--keep_telomeric_breakpoints",
            ],
        },
        {
            "name": "Output breakpoints",
            "options": [
                "--seq_window_size",
                "--min_supporting_reads",
                "--clustering_threshold",
            ],
        },
    ]
}


def filter_out_tents_intervals(input_tents: Tents, target_intervals: Intervals) -> None:
    to_remove = set()
    for i, input_tent in enumerate(input_tents):
        input_interval = Interval(input_tent.contig, input_tent.start, input_tent.end)
        for target_interval in target_intervals:
            if target_interval.overlaps_or_touches(input_interval):
                to_remove.add(i)
                break
    input_tents.drop_elements(to_remove)


def run_breakpoint_detection(
    detection_params: BreakpointDetectionParams,
    seq_regions: Intervals,
    threads: int,
    telomere_array_regions: Intervals,
) -> Tuple[PutativeBreakpoints, SoftclippedReads]:
    with mp.Pool(processes=threads) as pool:
        pooled_results = pool.starmap(
            find_breakpoint_foci,
            zip(
                it.repeat(detection_params),
                seq_regions,
            ),
        )
    all_reads_for_telomere_addition = list()
    all_foci = setup_breakpoint_tents()
    for foci_pool, read_pool in pooled_results:
        all_foci.extend(foci_pool)
        all_reads_for_telomere_addition.extend(read_pool)
    if (
        detection_params.breakpoint_type is BreakpointType.S2G
        and not detection_params.keep_telomeric_breakpoints
    ):
        filter_out_tents_intervals(all_foci, telomere_array_regions)
        all_reads_for_telomere_addition = filter_out_read_intervals(
            all_reads_for_telomere_addition, telomere_array_regions
        )
    foci_tsv = f"{detection_params.ofname_base}.tsv"
    with open(foci_tsv, "w") as ofstream:
        print(all_foci, file=ofstream)
    clustered_foci = cluster_breakpoint_foci(
        all_foci, tolerance=detection_params.clustering_threshold
    )
    putative_breakpoints = map(
        lambda cluster: cluster.find_peak_softclip_focus(), clustered_foci
    )
    putative_breakpoints = sorted(
        putative_breakpoints, key=lambda e: e.max_value, reverse=True
    )
    for m_f in putative_breakpoints:
        m_f.breakpoint_type = detection_params.breakpoint_type
    return putative_breakpoints, all_reads_for_telomere_addition


def write_breakpoint_bed(
    putative_breakpoints: PutativeBreakpoints, sample_name: str, odirname: Path
) -> None:
    breakpoint_bed = odirname / "breakpoint_locations.bed"
    with breakpoint_bed.open("w") as ofstream:
        for putative_breakpoint in putative_breakpoints:
            strand = putative_breakpoint.orientation.value
            breakpoint_name = f"Type:{putative_breakpoint.breakpoint_type};breakpoint_window:{putative_breakpoint.interval[0]}-{putative_breakpoint.interval[1]}"
            out_line = [
                putative_breakpoint.focus.contig,
                putative_breakpoint.focus.start,
                putative_breakpoint.focus.end,
                breakpoint_name,
                putative_breakpoint.max_value,
                strand,
                sample_name,
            ]
            ofstream.write("\t".join(map(str, out_line)) + "\n")


@click.command()
@click.argument("genome_fname", type=click.Path(exists=True))
@click.argument("bam_fname", type=click.Path(exists=True))
@click.argument("odirname")
@click.option(
    "--sample_name",
    type=str,
    help="The name of your sample; will appear in a dedicated column in all output TSV files",
    default=".",
    show_default=True,
)
@click.option("--seq_region", type=str, help=REGION_CLICK_HELP)
@click.option(
    "--bed",
    type=click.Path(exists=True),
    help="Path to bed of regions to analyse. Overrides 'seq_region'",
)
@click.option(
    "--telo_forward_seq",
    type=str,
    default=TELOMERE_SEQS["Nematoda"][Orientation.forward],
    help="The telomere sequence used by your organism. Please make sure this is provided in 'forward' orientation (i.e. 5'->3')",
    show_default=True,
)
@click.option(
    "--telo_array_size",
    type=int,
    default=10,
    help="Minimum number of telomeric repeats for a read to be recorded",
    show_default=True,
)
@click.option(
    "--telo_max_edit_distance",
    type=int,
    default=3,
    help="Maximum number of mutations allowed in the searched telomere array",
    show_default=True,
)
@click.option(
    "--clustering_threshold",
    type=int,
    default=5,
    help="Any identified breakpoints within this value (in bp) of each other will be merged. "
    "A larger threshold allows for more imprecise breakpoint locations",
    show_default=True,
)
@click.option(
    "--min_mapq",
    type=int,
    default=DEFAULT_MIN_MAPQ,
    help="Reads below this MAPQ will be filtered out",
    show_default=True,
)
@click.option(
    "--read_filter_flag",
    type=int,
    default=DEFAULT_READ_FILTER_FLAG,
    help=f"Reads with any of the component bitwise flags will be filtered out (see SAM specs for details)."
    f"   [default: {DEFAULT_READ_FILTER_FLAG} (reads with any of {DEFAULT_READ_FILTER_NAMES} are filtered out)]",
)
@click.option(
    "--min_supporting_reads",
    type=int,
    default=10,
    help="Minimum number of reads supporting a breakpoint",
    show_default=True,
)
@click.option(
    "--seq_window_size",
    type=int,
    default=350,
    help="Number of nucleotides to extract either side of each identified breakpoint",
    show_default=True,
)
@click.option(
    "--breakpoint_type",
    "-b",
    type=click.Choice(list(map(str, all_breakpoint_types)) + ["all"]),
    help="The type of breakpoint to look for. By default, looks for all",
    default="all",
)
@click.option(
    "--keep_telomeric_breakpoints",
    is_flag=True,
    help="Forces delfies to keep breakpoints occurring inside telomeric arrays. As these are often false positives, they are discarded by default.",
    show_default=True,
)
@click.option("--threads", type=int, default=1)
@click.help_option("--help", "-h")
@click.version_option(__version__, "--version", "-V")
def main(
    genome_fname,
    bam_fname,
    odirname,
    sample_name,
    seq_region,
    bed,
    telo_forward_seq,
    telo_array_size,
    telo_max_edit_distance,
    clustering_threshold,
    min_mapq,
    read_filter_flag,
    min_supporting_reads,
    seq_window_size,
    breakpoint_type,
    keep_telomeric_breakpoints,
    threads,
):
    """
    Looks for DNA Elimination breakpoints from a bam of reads aligned to a genome.

    odirname is the directory to store outputs in.
    """
    odirname = Path(odirname)
    odirname.mkdir(parents=True, exist_ok=True)
    ofname_base = odirname / "breakpoint_foci"
    bam_fstream = AlignmentFile(bam_fname)

    seq_regions: Intervals = list()
    if bed is not None:
        intervals = BedTool(bed)
        for interval in intervals:
            seq_regions.append(Interval.from_pybedtools_interval(interval))
    elif seq_region is not None:
        seq_regions.append(Interval.from_region_string(seq_region))
        threads = 1
    else:
        # Analyse the entire genome
        for contig in bam_fstream.references:
            seq_regions.append(Interval(contig))

    telomere_seqs = {
        Orientation.forward: telo_forward_seq,
        Orientation.reverse: rev_comp(telo_forward_seq),
    }

    clustering_threshold = max(clustering_threshold, 0)
    detection_params = BreakpointDetectionParams(
        bam_fname=bam_fname,
        sample_name=sample_name,
        telomere_seqs=telomere_seqs,
        telo_array_size=telo_array_size,
        max_edit_distance=telo_max_edit_distance,
        clustering_threshold=clustering_threshold,
        min_mapq=min_mapq,
        read_filter_flag=read_filter_flag,
        min_supporting_reads=min_supporting_reads,
        keep_telomeric_breakpoints=keep_telomeric_breakpoints,
    )

    try:
        breakpoint_types_to_analyse = [BreakpointType(breakpoint_type)]
    except ValueError:
        breakpoint_types_to_analyse = all_breakpoint_types

    identified_breakpoints = []
    searched_telo_unit = detection_params.telomere_seqs[Orientation.forward]
    searched_telo_array = searched_telo_unit * detection_params.telo_array_size
    interval_window_size = len(searched_telo_array)
    all_reads_for_telomere_addition = list()
    telomere_array_regions = find_all_occurrences_in_genome(
        searched_telo_array,
        genome_fname,
        seq_regions,
        interval_window_size,
    )
    for breakpoint_type_to_analyse in breakpoint_types_to_analyse:
        detection_params.breakpoint_type = breakpoint_type_to_analyse
        detection_params.ofname_base = (
            f"{ofname_base}{ID_DELIM}{breakpoint_type_to_analyse}"
        )
        if breakpoint_type_to_analyse is BreakpointType.G2S:
            # Restrict regions to analyse to those containing telomere arrays
            seq_regions = telomere_array_regions
        candidate_breakpoints, reads_for_telomere_addition = run_breakpoint_detection(
            detection_params, seq_regions, threads, telomere_array_regions
        )
        all_reads_for_telomere_addition.extend(reads_for_telomere_addition)
        identified_breakpoints.extend(candidate_breakpoints)

    write_breakpoint_bed(identified_breakpoints, sample_name, odirname)
    seq_window_size = max(seq_window_size, 1)
    write_breakpoint_sequences(
        genome_fname, identified_breakpoints, odirname, seq_window_size
    )

    telomere_additions = record_telomere_additions(
        all_reads_for_telomere_addition,
        len(telo_forward_seq),
        genome_fname,
        sample_name,
    )
    telomere_addition_ofname = odirname / "telomere_additions.tsv"
    with telomere_addition_ofname.open("w") as ofstream:
        print(telomere_additions, file=ofstream)


if __name__ == "__main__":
    main()
